@@ -31,6 +31,7 @@ from google.cloud.ndb import _options
 from google.cloud.ndb import _remote
 from google.cloud.ndb import _retry
 from google.cloud.ndb import tasklets
+from google.cloud.ndb import utils
 
 EVENTUAL = datastore_pb2.ReadOptions.EVENTUAL
 EVENTUAL_CONSISTENCY = EVENTUAL  # Legacy NDB
@@ -82,10 +83,13 @@ def make_call(rpc_name, request, retries=None, timeout=None):
 
     @tasklets.tasklet
     def rpc_call():
+        context = context_module.get_toplevel_context()
+
         call = method.future(request, timeout=timeout)
-        rpc = _remote.RemoteCall(call, "{}({})".format(rpc_name, request))
-        log.debug(rpc)
-        log.debug("timeout={}".format(timeout))
+        rpc = _remote.RemoteCall(call, rpc_name)
+        utils.logging_debug(log, rpc)
+        utils.logging_debug(log, "timeout={}", timeout)
+        utils.logging_debug(log, request)
 
         try:
             result = yield rpc
@@ -93,6 +97,8 @@ def make_call(rpc_name, request, retries=None, timeout=None):
             if isinstance(error, grpc.Call):
                 error = core_exceptions.from_grpc_error(error)
             raise error
+        finally:
+            context.rpc_time += rpc.elapsed_time
 
         raise tasklets.Return(result)
 
@@ -120,16 +126,13 @@ def lookup(key, options):
     """
     context = context_module.get_context()
     use_datastore = context._use_datastore(key, options)
-    in_transaction = bool(_get_transaction(options))
-    if use_datastore and in_transaction:
+    if use_datastore and options.transaction:
         use_global_cache = False
     else:
         use_global_cache = context._use_global_cache(key, options)
 
     if not (use_global_cache or use_datastore):
-        raise TypeError(
-            "use_global_cache and use_datastore can't both be False"
-        )
+        raise TypeError("use_global_cache and use_datastore can't both be False")
 
     entity_pb = _NOT_FOUND
     key_locked = False
@@ -139,25 +142,35 @@ def lookup(key, options):
         result = yield _cache.global_get(cache_key)
         key_locked = _cache.is_locked_value(result)
         if not key_locked:
-            if result is not None:
+            if result:
                 entity_pb = entity_pb2.Entity()
                 entity_pb.MergeFromString(result)
 
             elif use_datastore:
-                yield _cache.global_lock(cache_key)
-                yield _cache.global_watch(cache_key)
+                lock = yield _cache.global_lock_for_read(cache_key, result)
+                if lock:
+                    yield _cache.global_watch(cache_key, lock)
+
+                else:
+                    # Another thread locked or wrote to this key after the call to
+                    # _cache.global_get above. Behave as though the key was locked by
+                    # another thread and don't attempt to write our value below
+                    key_locked = True
 
     if entity_pb is _NOT_FOUND and use_datastore:
         batch = _batch.get_batch(_LookupBatch, options)
         entity_pb = yield batch.add(key)
 
         # Do not cache misses
-        if use_global_cache and not key_locked and entity_pb is not _NOT_FOUND:
-            expires = context._global_cache_timeout(key, options)
-            serialized = entity_pb.SerializeToString()
-            yield _cache.global_compare_and_swap(
-                cache_key, serialized, expires=expires
-            )
+        if use_global_cache and not key_locked:
+            if entity_pb is not _NOT_FOUND:
+                expires = context._global_cache_timeout(key, options)
+                serialized = entity_pb.SerializeToString()
+                yield _cache.global_compare_and_swap(
+                    cache_key, serialized, expires=expires
+                )
+            else:
+                yield _cache.global_unwatch(cache_key)
 
     raise tasklets.Return(entity_pb)
 
@@ -244,7 +257,7 @@ class _LookupBatch(object):
 
         # Process results, which are divided into found, missing, and deferred
         results = rpc.result()
-        log.debug(results)
+        utils.logging_debug(log, results)
 
         # For all deferred keys, batch them up again with their original
         # futures
@@ -252,9 +265,7 @@ class _LookupBatch(object):
             next_batch = _batch.get_batch(type(self), self.options)
             for key in results.deferred:
                 todo_key = key.SerializeToString()
-                next_batch.todo.setdefault(todo_key, []).extend(
-                    self.todo[todo_key]
-                )
+                next_batch.todo.setdefault(todo_key, []).extend(self.todo[todo_key])
 
         # For all missing keys, set result to _NOT_FOUND and let callers decide
         # how to handle
@@ -317,8 +328,7 @@ def get_read_options(options, default_read_consistency=None):
         ValueError: When ``read_consistency`` is set to ``EVENTUAL`` and there
             is a transaction.
     """
-    transaction = _get_transaction(options)
-
+    transaction = options.transaction
     read_consistency = options.read_consistency
 
     if transaction is None:
@@ -326,36 +336,13 @@ def get_read_options(options, default_read_consistency=None):
             read_consistency = default_read_consistency
 
     elif read_consistency is EVENTUAL:
-        raise ValueError(
-            "read_consistency must not be EVENTUAL when in transaction"
-        )
+        raise ValueError("read_consistency must not be EVENTUAL when in transaction")
 
     return datastore_pb2.ReadOptions(
         read_consistency=read_consistency, transaction=transaction
     )
 
 
-def _get_transaction(options):
-    """Get the transaction for a request.
-
-    If specified, this will return the transaction from ``options``. Otherwise,
-    it will return the transaction for the current context.
-
-    Args:
-        options (_options.ReadOptions): The options for the request. Only
-            ``transaction`` will have any bearing here.
-
-    Returns:
-        Union[bytes, NoneType]: The transaction identifier, or :data:`None`.
-    """
-    transaction = getattr(options, "transaction", None)
-    if transaction is None:
-        context = context_module.get_context()
-        transaction = context.transaction
-
-    return transaction
-
-
 @tasklets.tasklet
 def put(entity, options):
     """Store an entity in datastore.
@@ -375,25 +362,24 @@ def put(entity, options):
     use_global_cache = context._use_global_cache(entity.key, options)
     use_datastore = context._use_datastore(entity.key, options)
     if not (use_global_cache or use_datastore):
-        raise TypeError(
-            "use_global_cache and use_datastore can't both be False"
-        )
+        raise TypeError("use_global_cache and use_datastore can't both be False")
 
     if not use_datastore and entity.key.is_partial:
         raise TypeError("Can't store partial keys when use_datastore is False")
 
+    lock = None
     entity_pb = helpers.entity_to_protobuf(entity)
     cache_key = _cache.global_cache_key(entity.key)
     if use_global_cache and not entity.key.is_partial:
         if use_datastore:
-            yield _cache.global_lock(cache_key)
+            lock = yield _cache.global_lock_for_write(cache_key)
         else:
             expires = context._global_cache_timeout(entity.key, options)
             cache_value = entity_pb.SerializeToString()
             yield _cache.global_set(cache_key, cache_value, expires=expires)
 
     if use_datastore:
-        transaction = _get_transaction(options)
+        transaction = context.transaction
         if transaction:
             batch = _get_commit_batch(transaction, options)
         else:
@@ -405,8 +391,16 @@ def put(entity, options):
         else:
             key = None
 
-        if use_global_cache:
-            yield _cache.global_delete(cache_key)
+        if lock:
+            if transaction:
+
+                def callback():
+                    _cache.global_unlock_for_write(cache_key, lock).result()
+
+                context.call_on_transaction_complete(callback)
+
+            else:
+                yield _cache.global_unlock_for_write(cache_key, lock)
 
         raise tasklets.Return(key)
 
@@ -429,15 +423,15 @@ def delete(key, options):
     context = context_module.get_context()
     use_global_cache = context._use_global_cache(key, options)
     use_datastore = context._use_datastore(key, options)
+    transaction = context.transaction
 
     if use_global_cache:
         cache_key = _cache.global_cache_key(key)
 
     if use_datastore:
         if use_global_cache:
-            yield _cache.global_lock(cache_key)
+            lock = yield _cache.global_lock_for_write(cache_key)
 
-        transaction = _get_transaction(options)
         if transaction:
             batch = _get_commit_batch(transaction, options)
         else:
@@ -446,7 +440,18 @@ def delete(key, options):
         yield batch.delete(key)
 
     if use_global_cache:
-        yield _cache.global_delete(cache_key)
+        if transaction:
+
+            def callback():
+                _cache.global_unlock_for_write(cache_key, lock).result()
+
+            context.call_on_transaction_complete(callback)
+
+        elif use_datastore:
+            yield _cache.global_unlock_for_write(cache_key, lock)
+
+        else:
+            yield _cache.global_delete(cache_key)
 
 
 class _NonTransactionalCommitBatch(object):
@@ -801,7 +806,7 @@ def _process_commit(rpc, futures):
     #
     # https://github.com/googleapis/googleapis/blob/master/google/datastore/v1/datastore.proto#L241
     response = rpc.result()
-    log.debug(response)
+    utils.logging_debug(log, response)
 
     results_futures = zip(response.mutation_results, futures)
     for mutation_result, future in results_futures:
@@ -985,9 +990,7 @@ def _datastore_allocate_ids(keys, retries=None, timeout=None):
             :class:`google.cloud.datastore_v1.datastore_pb2.AllocateIdsResponse`
     """
     client = context_module.get_context().client
-    request = datastore_pb2.AllocateIdsRequest(
-        project_id=client.project, keys=keys
-    )
+    request = datastore_pb2.AllocateIdsRequest(project_id=client.project, keys=keys)
 
     return make_call("AllocateIds", request, retries=retries, timeout=timeout)
 
@@ -1045,9 +1048,7 @@ def _datastore_begin_transaction(read_only, retries=None, timeout=None):
         project_id=client.project, transaction_options=options
     )
 
-    return make_call(
-        "BeginTransaction", request, retries=retries, timeout=timeout
-    )
+    return make_call("BeginTransaction", request, retries=retries, timeout=timeout)
 
 
 @tasklets.tasklet
