@@ -17,29 +17,86 @@
 
 import collections
 import contextlib
+import itertools
+import os
 import six
 import threading
+import uuid
 
 from google.cloud.ndb import _eventloop
 from google.cloud.ndb import exceptions
 from google.cloud.ndb import key as key_module
-from google.cloud.ndb import tasklets
 
 
-__all__ = [
-    "AutoBatcher",
-    "Context",
-    "ContextOptions",
-    "get_context",
-    "TransactionOptions",
-]
+class _ContextIds:
+    """Iterator which generates a sequence of context ids.
 
+    Useful for debugging complicated interactions among concurrent processes and
+    threads.
 
-class _LocalState(threading.local):
-    """Thread local state."""
+    Each value in the sequence is a string that include the machine's "node", acquired
+    via `uuid.getnode()`, the current process id, and a sequence number which increases
+    monotonically starting from one in each process. The combination of all three is
+    sufficient to uniquely identify the context in which a particular piece of code is
+    being run. Each context, as it is created, is assigned the next id in this sequence.
+    The context id is used by `utils.logging_debug` to grant insight into where a debug
+    logging statement is coming from in a cloud evironment.
+    """
 
     def __init__(self):
-        self.context = None
+        self.prefix = "{}-{}-".format(uuid.getnode(), os.getpid())
+        self.counter = itertools.count(1)
+        self.lock = threading.Lock()
+
+    def __next__(self):
+        with self.lock:
+            sequence_number = next(self.counter)
+
+        return self.prefix + str(sequence_number)
+
+    next = __next__  # Python 2.7
+
+
+_context_ids = _ContextIds()
+
+
+try:  # pragma: NO PY2 COVER
+    import contextvars
+
+    class _LocalState:
+        """Thread local state."""
+
+        def __init__(self):
+            self._toplevel_context = contextvars.ContextVar(
+                "_toplevel_context", default=None
+            )
+            self._context = contextvars.ContextVar("_context", default=None)
+
+        @property
+        def context(self):
+            return self._context.get()
+
+        @context.setter
+        def context(self, value):
+            self._context.set(value)
+
+        @property
+        def toplevel_context(self):
+            return self._toplevel_context.get()
+
+        @toplevel_context.setter
+        def toplevel_context(self, value):
+            self._toplevel_context.set(value)
+
+
+except ImportError:  # pragma: NO PY3 COVER
+
+    class _LocalState(threading.local):
+        """Thread local state."""
+
+        def __init__(self):
+            self.context = None
+            self.toplevel_context = None
 
 
 _state = _LocalState()
@@ -73,6 +130,40 @@ def get_context(raise_context_error=True):
         raise exceptions.ContextError()
 
 
+def get_toplevel_context(raise_context_error=True):
+    """Get the current top level context.
+
+    This function should be called within a context established by
+    :meth:`google.cloud.ndb.client.Client.context`.
+
+    The toplevel context is the context created by the call to
+    :meth:`google.cloud.ndb.client.Client.context`. At times, this context will
+    be superceded by subcontexts, which are used, for example, during
+    transactions. This function will always return the top level context
+    regardless of whether one of these subcontexts is the current one.
+
+    Args:
+        raise_context_error (bool): If set to :data:`True`, will raise an
+            exception if called outside of a context. Set this to :data:`False`
+            in order to have it just return :data:`None` if called outside of a
+            context. Default: :data:`True`
+
+    Returns:
+        Context: The current context.
+
+    Raises:
+        .ContextError: If called outside of a context
+            established by :meth:`google.cloud.ndb.client.Client.context` and
+            ``raise_context_error`` is :data:`True`.
+    """
+    context = _state.toplevel_context
+    if context:
+        return context
+
+    if raise_context_error:
+        raise exceptions.ContextError()
+
+
 def _default_policy(attr_name, value_type):
     """Factory for producing default policies.
 
@@ -123,9 +214,7 @@ Defers to ``_use_global_cache`` on the Model class for the key's kind.
 See: :meth:`~google.cloud.ndb.context.Context.set_global_cache_policy`
 """
 
-_default_global_cache_timeout_policy = _default_policy(
-    "_global_cache_timeout", int
-)
+_default_global_cache_timeout_policy = _default_policy("_global_cache_timeout", int)
 """The default global cache timeout policy.
 
 Defers to ``_global_cache_timeout`` on the Model class for the key's kind.
@@ -145,6 +234,7 @@ See: :meth:`~google.cloud.ndb.context.Context.set_datastore_policy`
 _ContextTuple = collections.namedtuple(
     "_ContextTuple",
     [
+        "id",
         "client",
         "namespace",
         "eventloop",
@@ -154,6 +244,7 @@ _ContextTuple = collections.namedtuple(
         "cache",
         "global_cache",
         "on_commit_callbacks",
+        "transaction_complete_callbacks",
         "legacy_data",
     ],
 )
@@ -179,6 +270,7 @@ class _Context(_ContextTuple):
     def __new__(
         cls,
         client,
+        id=None,
         namespace=key_module.UNDEFINED,
         eventloop=None,
         batches=None,
@@ -191,11 +283,18 @@ class _Context(_ContextTuple):
         global_cache_timeout_policy=None,
         datastore_policy=None,
         on_commit_callbacks=None,
+        transaction_complete_callbacks=None,
         legacy_data=True,
+        retry=None,
+        rpc_time=None,
+        wait_time=None,
     ):
         # Prevent circular import in Python 2.7
         from google.cloud.ndb import _cache
 
+        if id is None:
+            id = next(_context_ids)
+
         if eventloop is None:
             eventloop = _eventloop.EventLoop()
 
@@ -213,6 +312,7 @@ class _Context(_ContextTuple):
 
         context = super(_Context, cls).__new__(
             cls,
+            id=id,
             client=client,
             namespace=namespace,
             eventloop=eventloop,
@@ -222,6 +322,7 @@ class _Context(_ContextTuple):
             cache=new_cache,
             global_cache=global_cache,
             on_commit_callbacks=on_commit_callbacks,
+            transaction_complete_callbacks=transaction_complete_callbacks,
             legacy_data=legacy_data,
         )
 
@@ -229,6 +330,7 @@ class _Context(_ContextTuple):
         context.set_global_cache_policy(global_cache_policy)
         context.set_global_cache_timeout_policy(global_cache_timeout_policy)
         context.set_datastore_policy(datastore_policy)
+        context.set_retry_state(retry)
 
         return context
 
@@ -239,7 +341,9 @@ class _Context(_ContextTuple):
         will be substituted.
         """
         fields = self._fields + tuple(self.__dict__.keys())
-        state = {name: getattr(self, name) for name in fields}
+        state = {
+            name: getattr(self, name) for name in fields if not name.startswith("_")
+        }
         state.update(kwargs)
         return type(self)(**state)
 
@@ -253,32 +357,19 @@ class _Context(_ContextTuple):
         """
         prev_context = _state.context
         _state.context = self
+        if not prev_context:
+            _state.toplevel_context = self
+            self.rpc_time = 0
+            self.wait_time = 0
         try:
             yield self
         finally:
             if prev_context:
                 prev_context.cache.update(self.cache)
+            else:
+                _state.toplevel_context = None
             _state.context = prev_context
 
-    @tasklets.tasklet
-    def _clear_global_cache(self):
-        """Clears the global cache.
-
-        Clears keys from the global cache that appear in the local context
-        cache. In this way, only keys that were touched in the current context
-        are affected.
-        """
-        # Prevent circular import in Python 2.7
-        from google.cloud.ndb import _cache
-
-        keys = [
-            _cache.global_cache_key(key._key)
-            for key in self.cache.keys()
-            if self._use_global_cache(key)
-        ]
-        if keys:
-            yield [_cache.global_delete(key) for key in keys]
-
     def _use_cache(self, key, options=None):
         """Return whether to use the context cache for this key."""
         flag = options.use_cache if options else None
@@ -481,6 +572,15 @@ class Context(_Context):
 
     set_memcache_timeout_policy = set_global_cache_timeout_policy
 
+    def get_retry_state(self):
+        return self._retry
+
+    def set_retry_state(self, state):
+        self._retry = state
+
+    def clear_retry_state(self):
+        self._retry = None
+
     def call_on_commit(self, callback):
         """Call a callback upon successful commit of a transaction.
 
@@ -506,6 +606,29 @@ class Context(_Context):
         else:
             callback()
 
+    def call_on_transaction_complete(self, callback):
+        """Call a callback upon completion of a transaction.
+
+        If not in a transaction, the callback is called immediately.
+
+        In a transaction, multiple callbacks may be registered and will be called once
+        the transaction completes, in the order in which they were registered. Callbacks
+        are called regardless of whether transaction is committed or rolled back.
+
+        If the callback raises an exception, it bubbles up normally.  This means: If the
+        callback is called immediately, any exception it raises will bubble up
+        immediately.  If the call is postponed until commit, remaining callbacks will be
+        skipped and the exception will bubble up through the transaction() call.
+        (However, the transaction is already committed or rolled back at that point.)
+
+        Args:
+            callback (Callable): The callback function.
+        """
+        if self.in_transaction():
+            self.transaction_complete_callbacks.append(callback)
+        else:
+            callback()
+
     def in_transaction(self):
         """Get whether a transaction is currently active.
 
@@ -515,6 +638,15 @@ class Context(_Context):
         """
         return self.transaction is not None
 
+    def in_retry(self):
+        """Get whether we are already in a retry block.
+
+        Returns:
+            bool: :data:`True` if currently in a retry block, otherwise
+                :data:`False`.
+        """
+        return self._retry is not None
+
     def memcache_add(self, *args, **kwargs):
         """Direct pass-through to memcache client."""
         raise exceptions.NoLongerImplementedError()
@@ -562,8 +694,19 @@ class ContextOptions(object):
 
 
 class TransactionOptions(object):
-    def __init__(self, *args, **kwargs):
-        raise exceptions.NoLongerImplementedError()
+    NESTED = 1  # join=False
+    MANDATORY = 2  # join=True
+    ALLOWED = 3  # join=True
+    INDEPENDENT = 4  # join=False
+
+    _PROPAGATION = frozenset((NESTED, MANDATORY, ALLOWED, INDEPENDENT))
+    _JOINABLE = frozenset((MANDATORY, ALLOWED))
+    _INT_TO_NAME = {
+        NESTED: "nested",
+        MANDATORY: "mandatory",
+        ALLOWED: "allowed",
+        INDEPENDENT: "independent",
+    }
 
 
 class AutoBatcher(object):
