@@ -179,12 +179,12 @@ _log = logging.getLogger(__name__)
 
 class PropertyOrder(object):
     """The sort order for a property name, to be used when ordering the
-       results of a query.
+    results of a query.
 
-       Args:
-           name (str): The name of the model property to use for ordering.
-           reverse (bool): Whether to reverse the sort order (descending)
-               or not (ascending). Default is False.
+    Args:
+        name (str): The name of the model property to use for ordering.
+        reverse (bool): Whether to reverse the sort order (descending)
+            or not (ascending). Default is False.
     """
 
     def __init__(self, name, reverse=False):
@@ -192,9 +192,7 @@ class PropertyOrder(object):
         self.reverse = reverse
 
     def __repr__(self):
-        return "PropertyOrder(name='{}', reverse={})".format(
-            self.name, self.reverse
-        )
+        return "PropertyOrder(name='{}', reverse={})".format(self.name, self.reverse)
 
     def __neg__(self):
         reverse = not self.reverse
@@ -257,8 +255,7 @@ class RepeatedStructuredPropertyPredicate(object):
                 subprop_name = prop_name.split(".", 1)[1]
                 if not subentities:
                     subentities = [
-                        {subprop_name: value}
-                        for value in prop_pb.array_value.values
+                        {subprop_name: value} for value in prop_pb.array_value.values
                     ]
                 else:
                     for subentity, value in zip(
@@ -308,9 +305,7 @@ class Parameter(ParameterizedThing):
     def __init__(self, key):
         if not isinstance(key, six.integer_types + six.string_types):
             raise TypeError(
-                "Parameter key must be an integer or string, not {}".format(
-                    key
-                )
+                "Parameter key must be an integer or string, not {}".format(key)
             )
         self._key = key
 
@@ -345,9 +340,7 @@ class Parameter(ParameterizedThing):
         """
         key = self._key
         if key not in bindings:
-            raise exceptions.BadArgumentError(
-                "Parameter :{} is not bound.".format(key)
-            )
+            raise exceptions.BadArgumentError("Parameter :{} is not bound.".format(key))
         value = bindings[key]
         used[key] = True
         return value
@@ -527,9 +520,7 @@ class ParameterNode(Node):
         if op not in _OPS:
             raise TypeError("Expected a valid operator, got {!r}".format(op))
         if not isinstance(param, ParameterizedThing):
-            raise TypeError(
-                "Expected a ParameterizedThing, got {!r}".format(param)
-            )
+            raise TypeError("Expected a ParameterizedThing, got {!r}".format(param))
         obj = super(ParameterNode, cls).__new__(cls)
         obj._prop = prop
         obj._op = op
@@ -657,9 +648,7 @@ class FilterNode(Node):
                     "in expected a list, tuple or set of values; "
                     "received {!r}".format(value)
                 )
-            nodes = [
-                FilterNode(name, _EQ_OP, sub_value) for sub_value in value
-            ]
+            nodes = [FilterNode(name, _EQ_OP, sub_value) for sub_value in value]
             if not nodes:
                 return FalseNode()
             if len(nodes) == 1:
@@ -730,9 +719,7 @@ class FilterNode(Node):
                 "to a single filter ({!r})".format(self._opsymbol)
             )
 
-        return _datastore_query.make_filter(
-            self._name, self._opsymbol, self._value
-        )
+        return _datastore_query.make_filter(self._name, self._opsymbol, self._value)
 
 
 class PostFilterNode(Node):
@@ -1191,9 +1178,7 @@ def _query_options(wrapped):
 
         if kwargs.get("keys_only"):
             if kwargs.get("projection"):
-                raise TypeError(
-                    "Cannot specify 'projection' with 'keys_only=True'"
-                )
+                raise TypeError("Cannot specify 'projection' with 'keys_only=True'")
             kwargs["projection"] = ["__key__"]
             del kwargs["keys_only"]
 
@@ -1281,8 +1266,11 @@ class QueryOptions(_options.ReadOptions):
             if not self.project:
                 self.project = context.client.project
 
-            if not self.namespace:
-                self.namespace = context.get_namespace()
+            if self.namespace is None:
+                if self.ancestor is None:
+                    self.namespace = context.get_namespace()
+                else:
+                    self.namespace = self.ancestor.namespace()
 
 
 class Query(object):
@@ -1304,6 +1292,10 @@ class Query(object):
             If not passed, uses the client's value.
         projection (list[Union[str, google.cloud.ndb.model.Property]]): The
             fields to return as part of the query results.
+        keys_only (bool): Return keys instead of entities.
+        offset (int): Number of query results to skip.
+        limit (Optional[int]): Maximum number of query results to return.
+            If not specified, there is no limit.
         distinct_on (list[str]): The field names used to group query
             results.
         group_by (list[str]): Deprecated. Synonym for distinct_on.
@@ -1326,6 +1318,9 @@ class Query(object):
         projection=None,
         distinct_on=None,
         group_by=None,
+        limit=None,
+        offset=None,
+        keys_only=None,
         default_options=None,
     ):
         # Avoid circular import in Python 2.7
@@ -1376,20 +1371,21 @@ class Query(object):
             projection = self._option("projection", projection)
             distinct_on = self._option("distinct_on", distinct_on)
             group_by = self._option("group_by", group_by)
+            limit = self._option("limit", limit)
+            offset = self._option("offset", offset)
+            keys_only = self._option("keys_only", keys_only)
 
         if ancestor is not None:
             if isinstance(ancestor, ParameterizedThing):
                 if isinstance(ancestor, ParameterizedFunction):
                     if ancestor.func != "key":
                         raise TypeError(
-                            "ancestor cannot be a GQL function"
-                            "other than Key"
+                            "ancestor cannot be a GQL function" "other than Key"
                         )
             else:
                 if not isinstance(ancestor, model.Key):
                     raise TypeError(
-                        "ancestor must be a Key; "
-                        "received {}".format(ancestor)
+                        "ancestor must be a Key; " "received {}".format(ancestor)
                     )
                 if not ancestor.id():
                     raise ValueError("ancestor cannot be an incomplete key")
@@ -1399,7 +1395,13 @@ class Query(object):
                 else:
                     project = ancestor.app()
                 if namespace is not None:
-                    if namespace != ancestor.namespace():
+                    # if namespace is the empty string, that means default
+                    # namespace, but after a put, if the ancestor is using
+                    # the default namespace, its namespace will be None,
+                    # so skip the test to avoid a false mismatch error.
+                    if namespace == "" and ancestor.namespace() is None:
+                        pass
+                    elif namespace != ancestor.namespace():
                         raise TypeError("ancestor/namespace mismatch")
                 else:
                     namespace = ancestor.namespace()
@@ -1430,6 +1432,9 @@ class Query(object):
         self.order_by = order_by
         self.project = project
         self.namespace = namespace
+        self.limit = limit
+        self.offset = offset
+        self.keys_only = keys_only
 
         self.projection = None
         if projection is not None:
@@ -1479,14 +1484,16 @@ class Query(object):
             args.append("filters=%r" % self.filters)
         if self.order_by is not None:
             args.append("order_by=%r" % self.order_by)
+        if self.limit is not None:
+            args.append("limit=%r" % self.limit)
+        if self.offset is not None:
+            args.append("offset=%r" % self.offset)
+        if self.keys_only is not None:
+            args.append("keys_only=%r" % self.keys_only)
         if self.projection:
-            args.append(
-                "projection=%r" % (_to_property_names(self.projection))
-            )
+            args.append("projection=%r" % (_to_property_names(self.projection)))
         if self.distinct_on:
-            args.append(
-                "distinct_on=%r" % (_to_property_names(self.distinct_on))
-            )
+            args.append("distinct_on=%r" % (_to_property_names(self.distinct_on)))
         if self.default_options is not None:
             args.append("default_options=%r" % self.default_options)
         return "%s(%s)" % (self.__class__.__name__, ", ".join(args))
@@ -1541,6 +1548,9 @@ class Query(object):
             default_options=self.default_options,
             projection=self.projection,
             distinct_on=self.distinct_on,
+            limit=self.limit,
+            offset=self.offset,
+            keys_only=self.keys_only,
         )
 
     def order(self, *props):
@@ -1571,6 +1581,9 @@ class Query(object):
             default_options=self.default_options,
             projection=self.projection,
             distinct_on=self.distinct_on,
+            limit=self.limit,
+            offset=self.offset,
+            keys_only=self.keys_only,
         )
 
     def analyze(self):
@@ -1649,6 +1662,9 @@ class Query(object):
             default_options=self.default_options,
             projection=self.projection,
             distinct_on=self.distinct_on,
+            limit=self.limit,
+            offset=self.offset,
+            keys_only=self.keys_only,
         )
 
     def _to_property_orders(self, order_by):
@@ -1698,8 +1714,6 @@ class Query(object):
         """Run a query, fetching results.
 
         Args:
-            limit (Optional[int]): Maximum number of results to fetch.
-                data:`None` or data:`0` indicates no limit.
             keys_only (bool): Return keys instead of entities.
             projection (list[Union[str, google.cloud.ndb.model.Property]]): The
                 fields to return as part of the query results.
@@ -1726,7 +1740,7 @@ class Query(object):
                 values for some of these arguments.
 
         Returns:
-            List([model.Model]): The query results.
+            List[Union[model.Model, key.Key]]: The query results.
         """
         return self.fetch_async(_options=kwargs["_options"]).result()
 
@@ -2005,7 +2019,7 @@ class Query(object):
         if futures:
             mapped_results = yield futures
         else:
-            mapped_results = []
+            mapped_results = ()
 
         raise tasklets.Return(mapped_results)
 
@@ -2167,7 +2181,6 @@ class Query(object):
         """
         return self.count_async(_options=kwargs["_options"]).result()
 
-    @tasklets.tasklet
     @_query_options
     @utils.keyword_only(
         offset=None,
@@ -2196,19 +2209,7 @@ class Query(object):
         # Avoid circular import in Python 2.7
         from google.cloud.ndb import _datastore_query
 
-        _options = kwargs["_options"]
-        options = _options.copy(projection=["__key__"], order_by=None)
-        results = _datastore_query.iterate(options, raw=True)
-        count = 0
-        limit = options.limit
-        while (yield results.has_next_async()):
-            count += 1
-            if limit and count == limit:
-                break
-
-            results.next()
-
-        raise tasklets.Return(count)
+        return _datastore_query.count(kwargs["_options"])
 
     @_query_options
     @utils.keyword_only(
@@ -2277,9 +2278,7 @@ class Query(object):
                 result returned, and `more` indicates whether there are
                 (likely) more results after that.
         """
-        return self.fetch_page_async(
-            None, _options=kwargs["_options"]
-        ).result()
+        return self.fetch_page_async(None, _options=kwargs["_options"]).result()
 
     @tasklets.tasklet
     @_query_options
@@ -2319,15 +2318,6 @@ class Query(object):
                     "that uses 'OR', '!=', or 'IN'."
                 )
 
-            post_filters = _options.filters._post_filters()
-            if post_filters:
-                raise TypeError(
-                    "Can't use 'fetch_page' or 'fetch_page_async' with a "
-                    "post-filter. (An in-memory filter.) This probably means "
-                    "you're querying a repeated structured property which "
-                    "requires post-filtering."
-                )
-
         iterator = _datastore_query.iterate(_options, raw=True)
         results = []
         cursor = None
@@ -2378,8 +2368,7 @@ def _to_property_names(properties):
             fixed.append(prop._name)
         else:
             raise TypeError(
-                "Unexpected property {}; "
-                "should be string or Property".format(prop)
+                "Unexpected property {}; " "should be string or Property".format(prop)
             )
     return fixed
 
